/* IMPORTANT
 * This snapshot file is auto-generated, but designed for humans.
 * It should be checked into source control and tracked carefully.
 * Re-generate by setting TAP_SNAPSHOT=1 and running tests.
 * Make sure to inspect the output below.  Do not ignore changes!
 */
'use strict'
exports[`test/ghsa-8qq5-rm4j-mr97.ts > TAP > warnings > must match snapshot 1`] = `
Array [
  Array [
    "TAR_ENTRY_INFO",
    "stripping / from absolute linkpath",
  ],
  Array [
    "TAR_ENTRY_ERROR",
    "ENOENT: no such file or directory, link",
  ],
  Array [
    "TAR_ENTRY_ERROR",
    "linkpath contains '..'",
  ],
  Array [
    "TAR_ENTRY_INFO",
    "stripping / from absolute linkpath",
  ],
  Array [
    "TAR_ENTRY_ERROR",
    "linkpath escapes extraction directory",
  ],
  Array [
    "TAR_ENTRY_INFO",
    "stripping / from absolute linkpath",
  ],
  Array [
    "TAR_ENTRY_INFO",
    "stripping c: from absolute linkpath",
  ],
  Array [
    "TAR_ENTRY_ERROR",
    "linkpath escapes extraction directory",
  ],
]
`
